/*
** Module   : ServerImpl.java
** Abstract : n class that implements the app server specific method and  attribute APIs.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130211 Created initial version.
** 002 CA  20130221 Added resourceDelete stub. setType was renamed to setResourceType. 
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CS  20130321 Added missing versions of connect() method.
** 005 CA  20130529 Added runtime support for appservers.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.net.SessionListener;
import com.goldencode.p2j.net.SessionManager;

/**
 * A class for implementing the Server methods and attributes.
 */
public class ServerImpl
extends HandleChain 
implements Server,
           AsyncRequestCountAttribute,
           FirstLastProcedureAttribute,
           SslServerAttribute,
           SubTypeAttribute,
           ConnectableServer,
           SessionListener
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(ServerImpl.class);

   /**
    * The appserver helper for this server, set after the {@link #connect} method is called. It
    * will be <code>null</code> if an web-service connection is estalbished.
    */
   private AppServerHelper helper = null;
   
   /**
    * Flag indicating if this server is connected to a web-service. If <code>false</code>, it is
    * not connected or connected to an appserver.
    */
   private boolean webService = false;
   
   /** The first procedure in this server's proxy procedure list. */
   private ProxyProcedureWrapper firstProcedure = null;
   
   /** The last procedure in this server's proxy procedure list. */
   private ProxyProcedureWrapper lastProcedure = null;

   /** Flag indicating if the server is valid. */
   private boolean valid = true;
   
   /**
    * Default constructor which initializes Server specific settings.
    */
   public ServerImpl()
   {
      setResourceType("SERVER");
      setUnknown(false);
   }
   
   /**
    * Check if the given handle refers a server resource.
    * 
    * @param    h
    *
    * @return   <code>true</code> if the handle's resource is a {@link ServerImpl server}.  
    */
   public static boolean isServer(handle h)
   {
      return h != null && !h.isUnknown() && (h.get() instanceof ServerImpl);
   }

   /**
    * Reports if this object is valid for use (has not been deleted).
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return valid;
   }

   /**
    * Returns the number of asynchronous requests.
    * 
    * @return   See above.
    */
   public integer getAsyncRequestCount()
   {
      // TODO Implement the actual logic.
      return new integer();
   }

   /**
    * Returns the subtype of the handle. 
    * <p>
    * For SERVER handles this represents the type of server to which this is connected. 
    * The possible values are "APPSERVER" and "WEBSERVICE".
    * 
    * @return   See above.
    */
   public character getSubType()
   {
      return new character(webService ? "WEBSERVICE" : "APPSERVER");
   }

   /**
    * Java implementation for getting the  SSL-SERVER-NAME  attribute  which contains the SSL 
    * server name if the connection is made as SSL.
    * 
    * @return   The SSL server name, if the connection is made as SSL or unknown
    *           otherwise
    */
   public character getSslServerName()
   {
      // TODO Implement the actual logic.
      return new character();
   }

   /**
    * Returns the first entry in the list of persistent procedures or procedure objects.
    * 
    * @return   See above.
    */
   public handle firstProcedure()
   {
      return new handle(firstProcedure);
   }

   /**
    * Returns the first entry in the list of persistent procedures or procedure objects.
    * 
    * @return   See above.
    */
   public handle lastProcedure()
   {
      return new handle(lastProcedure);
   }

   /**
    * This returns the id for the AppServer connection to this handle. For Web Services, the
    * returned  value will be an empty string.
    * 
    * @return   See above.
    */
   public character getClientConnectionId()
   {
      return helper == null ? new character("") : helper.getConnectionID();
   }

   /**
    * Returns a handle to the first asynchronous requests in the AppServer or WebService list of
    * asynchronous requests.
    * 
    * @return   See above.
    */
   public handle getFirstAsyncRequest()
   {
      // TODO Implement the actual logic.
      return null;
   }

   /**
    * Returns a handle to the last asynchronous requests in the AppServer or WebService list of
    * asynchronous requests.
    * 
    * @return   See above.
    */
   public handle getLastAsyncRequest()
   {
      // TODO Implement the actual logic.
      return null;
   }
   
   /**
    * This sends a stop condition to the AppServer or Web Service.
    * <p>
    * Depending on the type of server the stop condition is sent to, it may be handled  differently.
    * 
    * @return   <code>true</code> if CANCEL-REQUESTS() method returns <code>true</code>, 
    *           <code>false</code> otherwise.
    */
   public logical cancelRequests()
   {
      if (!_connected())
      {
         return new logical(false);
      }
      
      // TODO: 
      
      return new logical(true);
   }

   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed.
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public logical cancelRequestsAfter(integer timeSec)
   {
      if (!_connected())
      {
         return new logical(false);
      }

      // TODO Implement the actual logic.
      return cancelRequests();
   }
   
   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed.
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public logical cancelRequestsAfter(int64 timeSec)
   {
      // TODO Implement the actual logic.
      return cancelRequests();
   }
   
   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed.
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public logical cancelRequestsAfter(long timeSec)
   {
      return cancelRequestsAfter(new int64(timeSec));
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only 
    *           {@link character} values for this parameter.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(String options)
   {
      return connect(new character(options));
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only
    *           character values for this parameter.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options)
   {
      return connect(options, new character(), new character(), new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options, character user)
   {
      return connect(options, user, new character(), new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options, character user, character pwd)
   {
      return connect(options, user, pwd, new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * @param    serverInfo
    *           Other info to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code> otherwise
    */
   public logical connect(character options, character user, character pwd, character serverInfo)
   {
      if (_connected())
      {
         ErrorManager.displayError(5535, "Server is already connected", false);
         return new logical(false);
      }
      
      helper = null;

      Map<String, String> parms = parseOptions(options);
      if (parms == null)
      {
         return new logical(false);
      }
      
      boolean res = false;
      // check the -WSDL parameter, which is required for web services
      boolean webService = parms.containsKey("-WSDL");
      if (webService)
      {
         // the user, pwd and serverInfo params are ignored
         res = connectToWebService(parms);
      }
      else
      {
         res = connectToAppServer(parms, user, pwd, serverInfo);
      }
      
      if (!res)
      {
         return new logical(false);
      }

      // save the new state (only on successful connection)
      this.webService = webService;
      
      // register this server with the session
      Session sess = SessionManager.get().getSession();
      sess.addSessionListener(this);

      return new logical(true);
   }
   
   /**
    * Method declaration for the CONNECTED() method, supported by web-services and server handles.
    * This will return <code>true</code> if the web service is connected or <code>false</code>
    * otherwise.
    * 
    * @return   <code>true</code> if the server is connected or <code>false</code> otherwise.
    */
   public logical connected()
   {
      return new logical(_connected());
   }

   /**
    * Method declaration for the CONNECTED() method, supported by web-services and server handles.
    * This will return <code>true</code> if the web service is connected or <code>false</code>
    * otherwise.
    * 
    * @return   <code>true</code> if the server  is connected or <code>false</code> otherwise.
    */
   public boolean _connected()
   {
      return helper != null;
   }

   /**
    * Method declaration for the DISCONNECT() method supported by web-services and server handles.
    * This will attempt to disconnect the resource and returns <code>true</code> if succeeded or 
    * <code>false</code> otherwise.
    * 
    * @return   <code>true</code> if disconnect succeeded or <code>false</code> otherwise.
    */
   public logical disconnect()
   {
      if (!_connected())
      {
         ErrorManager.displayError(5537, "Server is not connected.", false);
         return new logical(false);
      }
      
      if (webService)
      {
         // TODO: disconnect from the web service
         return new logical(false);
      }
      else
      {
         helper.disconnect();
      }

      Session sess = SessionManager.get().getSession();
      sess.removeSessionListener(this);

      return new logical(true);
   }
   
   /**
    * This method is called when the session is ending.  It disconnects the current server, if
    * the session which created it is ending.
    *
    * @param    session
    *           The session that is ending.
    */
   public void terminate(Session session)
   {
      if (_connected())
      {
         disconnect();
      }
   }

   /**
    * Get the {@link AppServerHelper helper} for this server resource; will be not-null only if
    * this server is connected to an appserver.
    * 
    * @return   See above.
    */
   AppServerHelper getAppServerHelper()
   {
      return helper;
   }

   /**
    * Check if this server is connected to a web-service.
    * 
    * @return   See above.
    */
   boolean isWebService()
   {
      return webService;
   }

   /**
    * Set the first proxy procedure of this server's chain.
    * 
    * @param    proc
    *           The procedure instance.
    */
   void setFirstProcedure(ProxyProcedureWrapper proc)
   {
      this.firstProcedure = proc;
   }
   
   /**
    * Set the last proxy procedure of this server's chain.
    * 
    * @param    proc
    *           The procedure instance.
    */
   void setLastProcedure(ProxyProcedureWrapper proc)
   {
      this.lastProcedure = proc;
   }
   
   /**
    * Delete the resource. If the server is still connected, delete is not possible.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   protected boolean resourceDelete()
   {
      if (_connected())
      {
         final String msg = "SERVER %s is still connected.  Cannot DELETE";
         final String err = String.format(msg, getClientConnectionId().toStringMessage());
         ErrorManager.recordOrThrowError(5457, err);

         return false;
      }

      this.valid = false;
      return true;
   }

   /**
    * Connect to a web service, using the specified parameters.
    * 
    * @param    parms
    *           The connection parameters.
    *           
    * @return   <code>true</code> if the connection was possible.
    */
   private boolean connectToWebService(Map<String, String> parms)
   {
      helper = null;

      // TODO: implement this
      return true;
   }

   /**
    * Connect to an appserver using the specified parameters.
    * 
    * @param    parms
    *           The set of parameters to establish the connection. 4GL allows only character
    *           values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * @param    serverInfo
    *           Other info to send to the CONNECT procedure on the appserver side.
    *
    * @return   <code>true</code> if the connection was possible.
    */
   private boolean connectToAppServer(Map<String, String> parms,
                                      character           user,
                                      character           pwd,
                                      character           serverInfo)
   {
      String appService = parms.get("-AppService");
      String host = parms.get("-H");
      String serviceName = parms.get("-S");
      int port = 0;
      try
      {
         port = Integer.parseInt(serviceName);
      }
      catch (NumberFormatException e)
      {
         // ignore
      }
      String url = parms.get("-URL");
      if (url != null)
      {
         // get them from the URL
         String err = null;
         try
         {
            URL u = new URL(url);
            String protocol = u.getProtocol();
            if (!(protocol.equals("AppServer") ||
                  protocol.equals("http")      ||
                  protocol.equals("https")))
            {
               // if we are not connecting to an AppServer, than abort
               err = protocol;
            }
            else
            {
               host = u.getHost();
               port = u.getPort();
               serviceName = null;
            }
         }
         catch (MalformedURLException e)
         {
            err = url;
         }
         
         if (err != null)
         {
            ErrorManager.displayError(String.format("Invalid URL syntax: %s", err));
            ErrorManager.displayError(5468, "Application server connect failure", false);
            return false;
         }
      }
      boolean directConnect = parms.containsKey("-DirectConnect");
      
      String sessionModel = getString(parms, "-sessionModel", "Session-managed");
      boolean sessionFree = "Session-free".equalsIgnoreCase(sessionModel);
      // TODO: in session-free mode, is possible to establish more than one connection (to allow
      //       async requests). the following parameters are related to session-free mode 
      int connectionLifetime = getInt(parms, "-connectionLifetime", 300);
      int initialConnections = getInt(parms, "-initialConnections", 1);
      int maxConnections = getInt(parms, "-maxConnections", 0);
      // the -ns* options are not used by P2J

      try
      {
         this.helper = AppServerHelper.connectLegacyMode(appService, host, serviceName, port, 
                                                         directConnect, sessionFree,
                                                         user, pwd, serverInfo);
      }
      catch (NumberedException e)
      {
         ErrorManager.recordOrThrowError(e);
      }
      finally
      {
         if (!_connected())
         {
            final String msg = "Application server connect failure";
            ErrorManager.displayError(5468, msg, false);
   
            return false;
         }
      }
      
      return true;
   }

   /**
    * Get a string from the parameters map, using the specified key and defaulting to the 
    * specified value, if it does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   private String getString(Map<String, String> parms, String key, String def)
   {
      String val = parms.get(key);
      return val == null ? def : val;
   }

   /**
    * Parse the value for the specified key as an int, or default to the specified value if the
    * key does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   private int getInt(Map<String, String> parms, String key, int def)
   {
      String sval = parms.get(key);
      if (sval == null)
      {
         return def;
      }
   
      int val = 0;
      try
      {
         return Integer.parseInt(sval);
      }
      catch (NumberFormatException e)
      {
         val = 0;
      }
      
      return val;
   }

   /**
    * Parse and validate the given options; a map is returned with each parameters's value. If a
    * parameter doesn't have a value, the map will contain <code>null</code>.
    *  
    * @param    options
    *           The options from which the parameters will be extracted.
    *           
    * @return   The parameter-to-value map.
    */
   private Map<String, String> parseOptions(character options)
   {
      if (options.isUnknown())
      {
         // if unknown, it assumes it connects to an AppServer, and not WebService
         return new HashMap<String, String>();
      }
      
      final Set<String> knownAppServerOptions = new HashSet<String>();
      knownAppServerOptions.add("-AppService");
      knownAppServerOptions.add("-H");
      knownAppServerOptions.add("-S");
      knownAppServerOptions.add("-DirectConnect"); // no value
      knownAppServerOptions.add("-ssl"); // no value
      knownAppServerOptions.add("-nosessionreuse"); // no value
      knownAppServerOptions.add("-nohostverify"); // no value
      knownAppServerOptions.add("-pf");
      knownAppServerOptions.add("-AppServerKeepalive");
      knownAppServerOptions.add("-URL");
      knownAppServerOptions.add("-sessionModel");
      knownAppServerOptions.add("-connectionLifetime");
      knownAppServerOptions.add("-initialConnections");
      knownAppServerOptions.add("-maxConnections");
      knownAppServerOptions.add("-nsClientMaxPort");
      knownAppServerOptions.add("-nsClientMinPort");
      knownAppServerOptions.add("-nsClientPicklistExpiration");
      knownAppServerOptions.add("-nsClientPicklistSize");
      knownAppServerOptions.add("-nsClientPortRetry");
      knownAppServerOptions.add("-nsClientDelay");
      
      final Set<String> noValueOptions = new HashSet<String>();
      noValueOptions.add("-DirectConnect"); // no value
      noValueOptions.add("-ssl"); // no value
      noValueOptions.add("-nosessionreuse"); // no value
      noValueOptions.add("-nohostverify"); // no value
      
      final Set<String> knownWebServiceOptions = new HashSet<String>();
      knownWebServiceOptions.add("-WSDL");
      knownWebServiceOptions.add("-WSDLUserid");
      knownWebServiceOptions.add("-WSDLPassword");
      knownWebServiceOptions.add("-Service");
      knownWebServiceOptions.add("-ServiceNamespace");
      knownWebServiceOptions.add("-Port");
      knownWebServiceOptions.add("-Binding");
      knownWebServiceOptions.add("-BindingNamespace");
      knownWebServiceOptions.add("-SOAPEndpoint");
      knownWebServiceOptions.add("-SOAPEndpointUserid");
      knownWebServiceOptions.add("-SOAPEndpointPassword");
      knownWebServiceOptions.add("-TargetNamespace");
      knownWebServiceOptions.add("-connectionLifetime");
      knownWebServiceOptions.add("-maxConnections");
      knownWebServiceOptions.add("-nosessionreuse");
      knownWebServiceOptions.add("-nohostverify");
      knownWebServiceOptions.add("-pf");
      
      final Map<String, Integer> intParams = new HashMap<String, Integer>();
      intParams.put("-connectionLifetime", 0);
      intParams.put("-initialConnections", 0);
      intParams.put("-maxConnections", 0);
      
      HashMap<String, String> parms = new HashMap<String, String>();
      String s = options.getValue();
      
      String[] opts = s.split(" ");
      // determine the target; assume appServer
      Set<String> knownOptions = knownAppServerOptions;
      for (int i = 0; i < opts.length; i++)
      {
         if ("-WSDL".equals(opts))
         {
            knownOptions = knownWebServiceOptions;
            break;
         }
      }

      String arg = null;
      for (int i = 0; i < opts.length; i++)
      {
         String txt = opts[i];
         if (txt.length() == 0)
            continue;
         
         boolean err = false;
         if (txt.startsWith("-"))
         {
            if (!knownOptions.contains(txt))
            {
               err = true;
            }
            else
            {
               arg = txt;
            }
            
            // some options do not have a corresponding value
            if (noValueOptions.contains(arg))
            {
               parms.put(arg, null);
               arg = null;
            }
         }
         else
         {
            if (arg != null)
            {
               if (intParams.containsKey(arg))
               {
                  // ensure that int values can be parsed
                  int value = 0;
                  try
                  {
                     value = Integer.parseInt(txt);
                  }
                  catch (NumberFormatException e)
                  {
                     // ignore
                  }
                  int def = intParams.get(arg);
                  if (value <= def)
                  {
                     final String msg1 = "The %s parameter requires an argument greater than %d";
                     final String msg2 = "Unable to process parameters";
                     ErrorManager.displayError(1404, String.format(msg1, arg, def), false);
                     ErrorManager.displayError(5509, msg2, false);
                     return null;
                  }
               }

               parms.put(arg, txt);
            }
            else
            {
               err = true;
            }
            
            arg = null;
         }
         
         if (err)
         {
            final String format = "Could not recognise argument: %s";
            final String msg1 = String.format(format, txt);
            final String msg2 = "Unable to process parameters";
            ErrorManager.displayError(301, msg1);
            ErrorManager.displayError(5509, msg2, false);
            return null;
         }
      }
      
      if (parms.containsKey("-pf"))
      {
         String pf = parms.get("-pf");
         Map<String, String> pfParms = readPfFile(pf);
         if (pfParms == null)
         {
            final String msg = "Unable to open parameter file %s, errno 2";
            ErrorManager.recordOrShowError(1247, String.format(msg, pf), false);

            return null;
         }
         
         // the parameters from the file are not validated
         parms.putAll(pfParms);
      }
   
      return parms;
   }
   
   /**
    * Read the parameters configured in the given file.
    * 
    * @param    file
    *           The file containing the parameters.
    *           
    * @return   A map containing the value for each parameter or <code>null</code> if there were
    *           problems reading the file.
    */
   private Map<String, String> readPfFile(String file)
   {
      if (FileSystemOps.searchPath(file).isUnknown())
      {
         // file does not exist.
         return null;
      }
      
      StreamWrapper stream = new StreamWrapper();
      stream.assign(StreamFactory.openFileStream(file, false, false));
      Map<String, String> result = new HashMap<String, String>();
      try
      {
         String line = stream.readLn();
         while (line != null)
         {
            line = line.trim();
            if (line.startsWith("-"))
            {
               int sharpIdx = line.indexOf('#');
               if (sharpIdx >= 0)
               {
                  // drop trailing comment
                  line = line.substring(0, sharpIdx).trim();
               }
               
               // this is a configuration line
               String param = line;
               int spaceIdx = param.indexOf(' ');
               if (spaceIdx >= 0)
               {
                  param = param.substring(0, spaceIdx);
                  // drop the parameter, to leave only the value (if any);
                  line = line.substring(spaceIdx).trim();
               }
               
               String val = (line.length() == 0 ? null : line);
               
               result.put(param,  val);
            }
         }
      }
      catch (EOFException e)
      {
         // ignore
      }
      catch (IOException e)
      {
         if (LOG.isWarnEnabled())
         {
            LOG.warn("Problems reading parameter file " + file, e);
         }

         // read error
         return null;
      }
      catch (InterruptedException e)
      {
         if (LOG.isWarnEnabled())
         {
            LOG.warn("Problems reading parameter file " + file, e);
         }

         // other error
         return null;
      }

      return result;
   }
}
